# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['_vendor',
 '_vendor.networkx',
 '_vendor.networkx.algorithms',
 '_vendor.networkx.algorithms.approximation',
 '_vendor.networkx.algorithms.approximation.tests',
 '_vendor.networkx.algorithms.assortativity',
 '_vendor.networkx.algorithms.assortativity.tests',
 '_vendor.networkx.algorithms.bipartite',
 '_vendor.networkx.algorithms.bipartite.tests',
 '_vendor.networkx.algorithms.centrality',
 '_vendor.networkx.algorithms.centrality.tests',
 '_vendor.networkx.algorithms.coloring',
 '_vendor.networkx.algorithms.coloring.tests',
 '_vendor.networkx.algorithms.community',
 '_vendor.networkx.algorithms.community.tests',
 '_vendor.networkx.algorithms.components',
 '_vendor.networkx.algorithms.components.tests',
 '_vendor.networkx.algorithms.connectivity',
 '_vendor.networkx.algorithms.connectivity.tests',
 '_vendor.networkx.algorithms.flow',
 '_vendor.networkx.algorithms.flow.tests',
 '_vendor.networkx.algorithms.isomorphism',
 '_vendor.networkx.algorithms.isomorphism.tests',
 '_vendor.networkx.algorithms.link_analysis',
 '_vendor.networkx.algorithms.link_analysis.tests',
 '_vendor.networkx.algorithms.minors',
 '_vendor.networkx.algorithms.minors.tests',
 '_vendor.networkx.algorithms.node_classification',
 '_vendor.networkx.algorithms.node_classification.tests',
 '_vendor.networkx.algorithms.operators',
 '_vendor.networkx.algorithms.operators.tests',
 '_vendor.networkx.algorithms.shortest_paths',
 '_vendor.networkx.algorithms.shortest_paths.tests',
 '_vendor.networkx.algorithms.tests',
 '_vendor.networkx.algorithms.traversal',
 '_vendor.networkx.algorithms.traversal.tests',
 '_vendor.networkx.algorithms.tree',
 '_vendor.networkx.algorithms.tree.tests',
 '_vendor.networkx.classes',
 '_vendor.networkx.classes.tests',
 '_vendor.networkx.drawing',
 '_vendor.networkx.drawing.tests',
 '_vendor.networkx.generators',
 '_vendor.networkx.generators.tests',
 '_vendor.networkx.linalg',
 '_vendor.networkx.linalg.tests',
 '_vendor.networkx.readwrite',
 '_vendor.networkx.readwrite.json_graph',
 '_vendor.networkx.readwrite.json_graph.tests',
 '_vendor.networkx.readwrite.tests',
 '_vendor.networkx.testing',
 '_vendor.networkx.testing.tests',
 '_vendor.networkx.tests',
 '_vendor.networkx.utils',
 '_vendor.networkx.utils.tests',
 '_vendor.nr.io.graphviz',
 '_vendor.nr.python.environment',
 '_vendor.typeapi',
 'cli',
 'core',
 'core.executor',
 'lib',
 'test',
 'util']

package_data = \
{'': ['*']}

install_requires = \
['dill>=0.3.5', 'setuptools>=33.1.0', 'typing-extensions>=3.10.0.2']

extras_require = \
{'testing': ['pytest>=6.0.0']}

entry_points = \
{'console_scripts': ['kraken = kraken.cli.main:main'],
 'pytest11': ['kraken.test = kraken.test']}

setup_kwargs = {
    'name': 'kraken-core',
    'version': '0.8.4',
    'description': '',
    'long_description': "# kraken-core\n\n[![Python application](https://github.com/kraken-build/kraken-core/actions/workflows/python-package.yml/badge.svg)](https://github.com/kraken-build/kraken-core/actions/workflows/python-package.yml)\n[![PyPI version](https://badge.fury.io/py/kraken-core.svg)](https://badge.fury.io/py/kraken-core)\n\n__The Kraken build system.__\n\nKraken focuses on ease of use and simplicity to model complex task orchestration workflows.\n\nThe **`kraken-core`** packages provides a number of top-level packages under the _`kraken.`_ namespace. The packages provided by this\nproject are listed below. Other Python packages may provide additional top-level packages under this namespace.\n\n* `kraken.core` &ndash; Modeling of projects, tasks and the task graph.\n* `kraken.cli` &ndash; Provides the `kraken` cli.\n* `kraken.lib` &ndash; A small set of core utility tasks (nothing language specific).\n* `kraken.test` &ndash; Fixtures for integration-testing Kraken builds with Pytest.\n* `kraken.util` &ndash; Utilities and helpers.\n* `kraken._vendor` &ndash; Vendored third party libraries.\n\n__Reproducible build environments__\n\nWe recommend that you use `krakenw` to invoke Kraken builds instead of the `kraken` cli directly to ensure that you\nhave an isolated and reproducible build environment. Install the kraken wrapper cli via the `kraken-wrapper` package\nand define your build requirements at the top of your `.kraken.py` build script:\n\n```\n# ::requirements kraken-std>=0.3.0\nfrom kraken.std.cargo import cargo_build\ncargo_build()\n```\n\n__Vendored packages__\n\nWe're vendoring a number of third party packages for the purpose of reducing the burden of package resolution\nat installation time. This is particularly relevant for using Kraken in continuous integration systems to improve\nresolve times and PEX size.\n\n* `networkx`\n* `nr.io.graphviz`\n* `nr.python.environment`\n* `termcolor`\n* `types-termcolor`\n* `typeapi`\n",
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
